{-
  The ubiquitous Queue example!
-}

-- # import Property
-- # import QuickCheck
-- # import List
-- # import Bool.Proofs
-- # import List.Proofs

-- Helpers ----------------------------------------------------------------

null : (A:Set) -> List A -> Bool
null A []     = true
null A (x::xs) = false

head : (A:Set) -> List A -> A
head A (x::xs) = x
head A []     = epsilon A

tail : (A:Set) -> List A -> List A
tail A (x::xs) = xs
tail A []     = epsilon (List A)

orT : (x,y:Bool) -> Holds (x || y) ==> Holds x \/ Holds y
orT true y = fol-plugin()
orT false y = fol-plugin()

nullTrue : (A:Set) -> (xs:List A) -> Holds (null A xs) ==> Eq (List A) xs []
nullTrue A [] = fol-plugin()
nullTrue A (x::xs) = fol-plugin(List)

-- Queue specification ----------------------------------------------------

Queue = List

empty : (A:Set) -> Queue A
empty A = []

isEmpty : (A:Set) -> Queue A -> Bool
isEmpty = null

add : (A:Set) -> A -> Queue A -> Queue A
add A x q = append A q [x]

remove : (A:Set) -> Queue A -> Queue A
remove = tail

front : (A:Set) -> Queue A -> A
front = head

-- Queue implementation ---------------------------------------------------

QueueI : (A:Set) -> Set
QueueI A = {List A; List A}

flipQ : (A:Set) -> QueueI A -> QueueI A
flipQ A (f,b) = if (QueueI A) (null A f)
		    (reverse A b, [])
		    (f, b)

emptyI : (A:Set) -> QueueI A
emptyI A = ([], [])

isEmptyI : (A:Set) -> QueueI A -> Bool
isEmptyI A (f,_) = null A f

addI : (A:Set) -> A -> QueueI A -> QueueI A
addI A x (f,b) = flipQ A (f, x::b)

frontI : (A:Set) -> QueueI A -> A
frontI A (x::f,b)  = x
frontI A ([],b)	  = epsilon A

removeI : (A:Set) -> QueueI A -> QueueI A
removeI A (x::f,b) = flipQ A (f,b)
removeI A ([], b) = epsilon (QueueI A)

-- Retrieving the specification -------------------------------------------

retrieve : (A:Set) -> QueueI A -> Queue A
retrieve A (f,b) = append A f (reverse A b)

-- Generators -------------------------------------------------------------

arbQueue : Gen (Queue Nat)
arbQueue = arbList Nat arbNat

arbQueueI : Gen (QueueI Nat)
arbQueueI = arbSigma (List Nat) (\_ -> List Nat)
		(arbList Nat arbNat)
		(\_ -> arbList Nat arbNat)

-- Properties -------------------------------------------------------------

invariant : (A:Set) -> QueueI A -> Prop
invariant A (f,b) = Holds (null A b || not (null A f))

-- isEmptyI == isEmpty
PropIsEmpty : (A:Set) -> Prop
PropIsEmpty A = (q:QueueI A) ->
		  invariant A q ==>
		  Eq Bool (isEmptyI A q)
			  (isEmpty A (retrieve A q))

-- testIsEmpty : PropIsEmpty Nat
-- testIsEmpty = qc-plugin(arbQueueI)

propIsEmpty : (A:Set) -> PropIsEmpty A
-- propIsEmpty A ([],b) = assert-plugin()
-- propIsEmpty A (x::f,b) = assert-plugin()

propIsEmpty A ([],[])  = \_ -> reflexivity Bool true
propIsEmpty A ([],x::xs)  = \p -> fol-plugin(p)

-- fol-plugin( Holds, Bool, propHoldsOr, rev
-- 				  , nullTrue A, propHoldsTrue)
propIsEmpty A (x::f,b) = \_ -> reflexivity Bool false

