{-| The parser produces concrete syntax. This is translated to "InternalSyntax"
    before type checking.  

The declarations of the concrete syntax differ considerably from the
ones of the internal syntax.  Also, concrete syntax has explicit
parentheses and infix operators in a left associative way.  The
translation to internal syntax, reassociates infix expressions
according to their fixities.  

Fixeties: Should be connected to the declaration of an identifier.
Their scope should be the scope of the identifier.  In the extreme,
this could be taken to
@
  \ (infixl 6 +) -> x + y
@
But I am not sure this is necessary.
-}

module Syntax.Concrete
	( module Syntax.Name
	, module Syntax.Concrete
	)
    where

import Syntax.Name
import Syntax.Position

-- Concrete syntax --------------------------------------------------------

type Type = Term

data Term   = App Term Term		-- ^ . @t t'@
	    | Name Name			-- ^ . @x@
	    | Lam Name Term		-- ^ . @\ x -> t@
	    | Star Universe		-- ^ . @#n | Set | Type@
	    | Prop			-- ^ . @Prop@ (if universe polymorphism)
	    | Pi Name Term Term		-- ^ . @(x::t) -> t'@
            | CallPlugin Name [Term]	-- ^ . @name-plugin(t1,...,tn)@
	    | Let [Decl] Term		-- ^ . @let ds in t@
	    | Paren Term		-- ^ . @( t )@
	    | InfixApp Term Name Term	-- ^ . @t `x` t' | t + t'@

data Universe = Univ Int

type Program	= [Decl]

-- | We currently have two forms of definitions. Normal definitions and
--   definitions of new constants.

data Decl	= TypeSig Name Type
		    -- ^ . @x :: t@
		| FunClause Name Term
		    -- ^ . @f = e@
		| Infix Assoc Precedence [Name]
                    -- ^ . @infix 1 == | infixl 2 `app` | infixr 3 =>@

data Assoc	= NonAssoc | LeftAssoc | RightAssoc

{- | Precedences are positive integers.  
Juxtaposition has precedence +infinity. The default precedence is 9. -}
type Precedence = Int
type Fixity	= (Assoc, Precedence)

defaultFixity :: Fixity
defaultFixity = (LeftAssoc, 9)

