{-| Reconstructing a term from a value. Useful for plugins
    that wants to generate terms that are not guaranteed to
    be type correct. Does not work for 'VTop'.
-}
module TypeChecker.ValueToTerm where

import Syntax.Internal
import Syntax.Values
import Syntax.Position

import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Context
import TypeChecker.Context
import TypeChecker.UnfoldConstants
import TypeChecker.Errors
import TypeChecker.Abstractions

import Utils.Names
import Utils.Monads

valueToTerm :: Value -> TC Term
valueToTerm = valueToTerm' (\_ _ -> return $ Meta noRange DontCare)

-- | Parameterised by how to handle meta variables
valueToTerm' :: (MetaRef -> [Term] -> TC Term) -> Value -> TC Term
valueToTerm' meta v =
    do	v' <- unfoldConstants v
	case vView v' of
	    VBId n  ->
		do  gamma <- getContext
		    return $ Var noRange $ [ x | (x,Abstract _) <- gamma ] !! n
		    
	    VId x	 -> return $ Var noRange x
	    VDef x _	 -> return $ Def noRange x
	    VCon x _ _ _ -> return $ Con noRange x
	    VPi v f	 ->
		do  a <- vTT (unEl v)
		    x <- fresh (absName f)
		    b <- vTT $ unEl $ absApp f (vId x)
		    return $ Pi noRange (VarP x) a b
	    VLam f			->
		do  x <- fresh (absName f)
		    b <- vTT $ absApp f (vId x)
		    return $ Lam noRange (VarP x) b
	    VSort (VSUniv u)	-> return $ Hash noRange u
	    VSort VSISet	-> return $ ISet noRange
	    VSort VSProp	-> return $ Prop noRange
	    VTop		-> failure cannotReconstructTopError
	    VApp v1 v2		->
		case vAppView v' of
		    VAVDef x h vs	      -> hApp h (Def noRange x) <$> mapM vTT vs
		    VAVCon x _ h _ vs	      -> hApp h (Con noRange x) <$> mapM vTT vs
		    VAVMeta m vs	      -> meta m =<< mapM vTT vs
		    VAVProj v (VField h x) vs ->
			do  e  <- vTT v
			    es <- mapM vTT vs
			    return $ hApp h (Project noRange e (Field noRange x)) es
		    _		    -> App <$> vTT v1 <*> vTT v2
	    VProj v p		->
		Project noRange <$> vTT v <*> return (projToTerm p)
	    VPair v w 		->
		Pair noRange <$> vTT v <*> vTT w
	    VSigma w f   ->
		do  a   <- vTT (unEl w)
		    z   <- fresh (absName f)
		    b   <- vTT $ unEl $ absApp f (vId z)
		    return $ Sigma noRange z a b
	    VMeta m	-> meta m []
	    VStruct xs	-> Struct noRange <$> mapM vtt xs
		where
		    vtt (x,(h,v)) =
			do  x' <- fresh x
			    (hs,w) <- hiding h v
			    e  <- vTT w
			    return (x',hs,[],e)

		    hiding 0 v = return ([],v)
		    hiding n v =
			do  x <- name $ vView v
			    (hs,w) <- hiding (n - 1) $ v `vApply` vId x
			    return (VarP x:hs,w)

		    name (VLam f) = fresh (absName f)
		    name _	  = fresh "h"
    where
	vTT = valueToTerm' meta
	projToTerm VFst		= Proj noRange 1
	projToTerm VSnd		= Proj noRange 2
	projToTerm (VField _ x) = Field noRange x

	hApp n e es = foldl App (foldl HApp e es0) es1
	    where
		(es0,es1) = splitAt n es


